@@ -1,20 +1,6 @@
-#     Copyright 2023, Kay Hayen, mailto:kay.hayen@gmail.com
-#
-#     Part of "Nuitka", an optimizing Python compiler that is compatible and
-#     integrates with CPython, but also works on its own.
-#
-#     Licensed under the Apache License, Version 2.0 (the "License");
-#     you may not use this file except in compliance with the License.
-#     You may obtain a copy of the License at
-#
-#        http://www.apache.org/licenses/LICENSE-2.0
-#
-#     Unless required by applicable law or agreed to in writing, software
-#     distributed under the License is distributed on an "AS IS" BASIS,
-#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-#     See the License for the specific language governing permissions and
-#     limitations under the License.
-#
+#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file
+
+
 """ Recursion into other modules.
 
 """
@@ -24,44 +10,58 @@ import os
 
 from nuitka import ModuleRegistry, Options
 from nuitka.Errors import NuitkaForbiddenImportEncounter
-from nuitka.importing import ImportCache, Importing, StandardLibrary
+from nuitka.freezer.ImportDetection import (
+    detectEarlyImports,
+    detectStdlibAutoInclusionModules,
+)
+from nuitka.importing import ImportCache, StandardLibrary
 from nuitka.ModuleRegistry import addUsedModule, getRootTopModule
 from nuitka.pgo.PGO import decideInclusionFromPGO
 from nuitka.plugins.Plugins import Plugins
 from nuitka.PythonVersions import python_version
 from nuitka.Tracing import recursion_logger
 from nuitka.utils.FileOperations import listDir
-from nuitka.utils.Importing import getSharedLibrarySuffixes
+from nuitka.utils.Importing import (
+    getExtensionModuleSuffixes,
+    getPackageDirFilename,
+)
 from nuitka.utils.ModuleNames import ModuleName
 
+from .Importing import (
+    getModuleNameAndKindFromFilename,
+    isPackageDir,
+    locateModule,
+    warnAboutNotFoundImport,
+)
 
-def _recurseTo(module_name, module_filename, module_kind):
+
+def _recurseTo(module_name, module_filename, module_kind, reason):
     from nuitka.tree import Building
 
-    module, is_added = Building.buildModule(
-        module_filename=module_filename,
+    module = Building.buildModule(
         module_name=module_name,
+        module_kind=module_kind,
+        module_filename=module_filename,
+        reason=reason,
         source_code=None,
         is_top=False,
         is_main=False,
-        is_extension=module_kind == "extension",
         is_fake=False,
         hide_syntax_error=True,
     )
 
     ImportCache.addImportedModule(module)
 
-    return module, is_added
+    return module
 
 
 def recurseTo(
-    signal_change,
     module_name,
     module_filename,
     module_kind,
-    using_module_name,
     source_ref,
     reason,
+    using_module_name,
 ):
     try:
         module = ImportCache.getImportedModuleByNameAndPath(
@@ -77,26 +77,54 @@ def recurseTo(
             module_kind=module_kind,
             using_module_name=using_module_name,
             source_ref=source_ref,
+            reason=reason,
         )
 
-        module, added_flag = _recurseTo(
+        module = _recurseTo(
             module_name=module_name,
             module_filename=module_filename,
             module_kind=module_kind,
+            reason=reason,
         )
 
-        if added_flag and signal_change is not None:
-            signal_change("new_code", module.getSourceReference(), reason)
-
     return module
 
 
 _recursion_decision_cache = {}
 
 
+def getRecursionDecisions():
+    """Access to recursion decisions, intended only for reporting."""
+    return _recursion_decision_cache
+
+
 def decideRecursion(
     using_module_name, module_filename, module_name, module_kind, extra_recursion=False
 ):
+    package_part, _remainder = module_name.splitModuleBasename()
+
+    if package_part is not None:
+        (
+            _package_part,
+            package_filename,
+            package_module_kind,
+            package_finding,
+        ) = locateModule(module_name=package_part, parent_package=None, level=0)
+        assert _package_part == package_part
+
+        # For bad decisions, this may already not work.
+        if package_finding != "not-found":
+            package_decision, package_reason = decideRecursion(
+                using_module_name=using_module_name,
+                module_filename=package_filename,
+                module_name=package_part,
+                module_kind=package_module_kind,
+                extra_recursion=extra_recursion,
+            )
+
+            if package_decision is False:
+                return package_decision, package_reason
+
     key = using_module_name, module_filename, module_name, module_kind, extra_recursion
 
     if key not in _recursion_decision_cache:
@@ -108,6 +136,16 @@ def decideRecursion(
             extra_recursion,
         )
 
+        # If decided true, give the plugins a chance to e.g. add more hard
+        # module information, this indicates tentatively, that a module might
+        # get used, but it may also not happen at all.
+        if _recursion_decision_cache[key][0]:
+            Plugins.onModuleUsageLookAhead(
+                module_name=module_name,
+                module_filename=module_filename,
+                module_kind=module_kind,
+            )
+
     return _recursion_decision_cache[key]
 
 
@@ -119,7 +157,26 @@ def _decideRecursion(
     if module_name == "__main__":
         return False, "Main program is not followed to a second time."
 
-    # In -m mode, when including the package, do not duplicate main program.
+    if using_module_name == "__main__" and Options.isMultidistMode():
+        return True, "Main program multidist entry points."
+
+    if module_kind == "extension" and not Options.isStandaloneMode():
+        return False, "Extension modules cannot be inspected."
+
+    if module_kind == "built-in":
+        return False, "Built-in modules cannot be inspected."
+
+    if module_name in detectEarlyImports():
+        return True, "Technically required for CPython library startup."
+
+    is_stdlib = module_filename is not None and StandardLibrary.isStandardLibraryPath(
+        module_filename
+    )
+
+    if is_stdlib and module_name in detectStdlibAutoInclusionModules():
+        return True, "Including as part of the non-excluded parts of standard library."
+
+    # In '-m' mode, when including the package, do not duplicate main program.
     if (
         Options.hasPythonFlagPackageMode()
         and not Options.shallMakeModule()
@@ -128,25 +185,63 @@ def _decideRecursion(
         if module_name.getPackageName() == getRootTopModule().getRuntimePackageValue():
             return False, "Main program is already included in package mode."
 
-    plugin_decision = Plugins.onModuleEncounter(
+    plugin_decision, deciding_plugins = Plugins.onModuleEncounter(
         using_module_name=using_module_name,
         module_filename=module_filename,
         module_name=module_name,
         module_kind=module_kind,
     )
 
+    # For checks, we do not consider anti-bloat plugin to be as important, to
+    # cause a conflict with the user choices. We could make an extra attribute
+    # out of that, but it seems overkill for now.
+    deciding_plugins = [
+        deciding_plugin
+        for deciding_plugin in deciding_plugins
+        if deciding_plugin.plugin_name != "anti-bloat"
+    ]
+
+    no_case, reason = module_name.matchesToShellPatterns(
+        patterns=Options.getShallFollowInNoCase()
+    )
+
+    if no_case:
+        if plugin_decision and plugin_decision[0]:
+            deciding_plugins[0].sysexit(
+                "Conflict between user and plugin decision for module '%s'."
+                % module_name
+            )
+
+        return False, "Module %s instructed by user to not follow to." % reason
+
+    any_case, reason = module_name.matchesToShellPatterns(
+        patterns=Options.getShallFollowModules()
+    )
+
+    if any_case:
+        if plugin_decision and not plugin_decision[0] and deciding_plugins:
+            deciding_plugins[0].sysexit(
+                "Conflict between user and plugin decision for module '%s'."
+                % module_name
+            )
+
+        return True, "Module %s instructed by user to follow to." % reason
+
     if plugin_decision is not None:
         return plugin_decision
 
-    if module_kind == "extension":
-        if Options.isStandaloneMode():
-            return True, "Extension module needed for standalone mode."
-        else:
-            return False, "Extension module cannot be inspected."
+    if Options.shallMakePackage():
+        if module_name.hasNamespace(getRootTopModule().getFullName()):
+            return True, "Submodule of compiled package."
+
+    if extra_recursion:
+        return True, "Lives in user provided directory."
+
+    if module_kind == "extension" and Options.isStandaloneMode():
+        return True, "Extension module needed for standalone mode."
 
     # PGO decisions are not overruling plugins, but all command line options, they are
     # supposed to be applied already.
-    is_stdlib = StandardLibrary.isStandardLibraryPath(module_filename)
 
     if not is_stdlib or Options.shallFollowStandardLibrary():
         # TODO: Bad placement of this function or should PGO also know about
@@ -154,7 +249,12 @@ def _decideRecursion(
         from nuitka.tree.Building import decideCompilationMode
 
         if (
-            decideCompilationMode(is_top=False, module_name=module_name, for_pgo=True)
+            decideCompilationMode(
+                is_top=False,
+                module_name=module_name,
+                module_filename=module_filename,
+                for_pgo=True,
+            )
             == "compiled"
         ):
             pgo_decision = decideInclusionFromPGO(
@@ -165,38 +265,21 @@ def _decideRecursion(
             if pgo_decision is not None:
                 return pgo_decision, "PGO based decision"
 
-    no_case, reason = module_name.matchesToShellPatterns(
-        patterns=Options.getShallFollowInNoCase()
-    )
-
-    if no_case:
-        return (False, "Module %s instructed by user to not follow to." % reason)
-
-    any_case, reason = module_name.matchesToShellPatterns(
-        patterns=Options.getShallFollowModules()
-    )
-
-    if any_case:
-        return (True, "Module %s instructed by user to follow to." % reason)
-
-    if extra_recursion:
-        return (True, "Lives in plug-in directory.")
-
-    if is_stdlib and Options.shallFollowStandardLibrary():
-        return (True, "Instructed by user to follow to standard library.")
+    if (
+        is_stdlib
+        and not Options.isStandaloneMode()
+        and not Options.shallFollowStandardLibrary()
+    ):
+        return (
+            False,
+            "Not following into stdlib unless standalone or requested to follow into stdlib.",
+        )
 
     if Options.shallFollowAllImports():
-        if is_stdlib:
-            if StandardLibrary.isStandardLibraryNoAutoInclusionModule(module_name):
-                return (
-                    True,
-                    "Instructed by user to follow all modules, including non-automatic standard library modules.",
-                )
-        else:
-            return (
-                True,
-                "Instructed by user to follow to all non-standard library modules.",
-            )
+        return (
+            True,
+            "Instructed by user to follow to all modules.",
+        )
 
     if Options.shallFollowNoImports():
         return (None, "Instructed by user to not follow at all.")
@@ -217,8 +300,9 @@ def isSameModulePath(path1, path2):
     return os.path.abspath(path1) == os.path.abspath(path2)
 
 
-def _addIncludedModule(module):
-    # Many branches, for the decision is very complex, pylint: disable=too-many-branches
+def _addIncludedModule(module, package_only):
+    # Many branches, for the decision is very complex
+    # pylint: disable=too-many-branches
 
     if Options.isShowInclusion():
         recursion_logger.info(
@@ -240,8 +324,8 @@ def _addIncludedModule(module):
 
             package_dir = package_filename
 
-            # Only include it, if it contains actual modules, which will
-            # recurse to this one and find it again.
+            # Only include it, if it contains actual modules, they will recurse
+            # to this one as a parent and find it again.
         else:
             package_dir = os.path.dirname(package_filename)
 
@@ -251,16 +335,38 @@ def _addIncludedModule(module):
         if Options.isShowInclusion():
             recursion_logger.info("Package directory '%s'." % package_dir)
 
-        for sub_path, sub_filename in listDir(package_dir):
-            if sub_filename in ("__init__.py", "__pycache__"):
-                continue
+        if not package_only:
+            for sub_path, sub_filename in listDir(package_dir):
+                if sub_filename in ("__init__.py", "__pycache__"):
+                    continue
 
-            if Importing.isPackageDir(sub_path) and not os.path.exists(
-                sub_path + ".py"
-            ):
-                checkPluginSinglePath(sub_path, module_package=module.getFullName())
-            elif sub_path.endswith(".py"):
-                checkPluginSinglePath(sub_path, module_package=module.getFullName())
+                if isPackageDir(sub_path) and not os.path.exists(sub_path + ".py"):
+                    scanPluginSinglePath(
+                        sub_path,
+                        module_package=module.getFullName(),
+                        package_only=False,
+                    )
+                elif sub_filename.endswith(".py"):
+                    if os.path.isdir(sub_path[:-3]):
+                        if getPackageDirFilename(sub_path[:-3]) is not None:
+                            continue
+
+                    scanPluginSinglePath(
+                        sub_path,
+                        module_package=module.getFullName(),
+                        package_only=False,
+                    )
+                else:
+                    for suffix in getExtensionModuleSuffixes():
+                        if (
+                            sub_filename.endswith(suffix)
+                            and "." not in sub_filename[: -len(suffix)]
+                        ):
+                            scanPluginSinglePath(
+                                sub_path,
+                                module_package=module.getFullName(),
+                                package_only=False,
+                            )
 
     elif module.isCompiledPythonModule() or module.isUncompiledPythonModule():
         ModuleRegistry.addRootModule(module)
@@ -271,7 +377,7 @@ def _addIncludedModule(module):
         assert False, module
 
 
-def checkPluginSinglePath(plugin_filename, module_package):
+def scanPluginSinglePath(plugin_filename, module_package, package_only):
     # The importing wants these to be unique.
     plugin_filename = os.path.abspath(plugin_filename)
 
@@ -281,9 +387,7 @@ def checkPluginSinglePath(plugin_filename, module_package):
             % (plugin_filename, module_package)
         )
 
-    module_name, module_kind = Importing.getModuleNameAndKindFromFilename(
-        plugin_filename
-    )
+    module_name, module_kind = getModuleNameAndKindFromFilename(plugin_filename)
 
     module_name = ModuleName.makeModuleNameInPackage(module_name, module_package)
 
@@ -298,7 +402,7 @@ the compiled result, and therefore asking to include them makes no sense.
         )
 
     if module_kind is not None:
-        decision, reason = decideRecursion(
+        decision, decision_reason = decideRecursion(
             using_module_name=None,
             module_filename=plugin_filename,
             module_name=module_name,
@@ -308,17 +412,16 @@ the compiled result, and therefore asking to include them makes no sense.
 
         if decision:
             module = recurseTo(
-                signal_change=None,
                 module_filename=plugin_filename,
                 module_name=module_name,
                 module_kind=module_kind,
-                using_module_name=None,
                 source_ref=None,
-                reason=reason,
+                reason="command line",
+                using_module_name=None,
             )
 
             if module:
-                _addIncludedModule(module)
+                _addIncludedModule(module=module, package_only=package_only)
             else:
                 recursion_logger.warning(
                     "Failed to include module from '%s'." % plugin_filename
@@ -326,33 +429,45 @@ the compiled result, and therefore asking to include them makes no sense.
         else:
             recursion_logger.warning(
                 "Not allowed to include module '%s' due to '%s'."
-                % (module_name, reason)
+                % (module_name, decision_reason)
             )
 
 
-def checkPluginPath(plugin_filename, module_package):
+def scanPluginPath(plugin_filename, module_package):
     if Options.isShowInclusion():
         recursion_logger.info(
-            "Checking top level inclusion path '%s' '%s'."
+            "Checking top level inclusion path '%s' for package '%s'."
             % (plugin_filename, module_package)
         )
 
     # Files and package directories are handled here.
-    if os.path.isfile(plugin_filename) or Importing.isPackageDir(plugin_filename):
-        checkPluginSinglePath(plugin_filename, module_package=module_package)
+    if os.path.isfile(plugin_filename) or isPackageDir(plugin_filename):
+        scanPluginSinglePath(
+            plugin_filename,
+            module_package=module_package,
+            package_only=False,
+        )
     # This effectively only covers files known to not be packages due to name
     # or older Python version.
     elif os.path.isdir(plugin_filename):
         for sub_path, sub_filename in listDir(plugin_filename):
             assert sub_filename != "__init__.py"
 
-            if Importing.isPackageDir(sub_path) or sub_path.endswith(".py"):
-                checkPluginSinglePath(sub_path, module_package=None)
+            if isPackageDir(sub_path) or sub_path.endswith(".py"):
+                scanPluginSinglePath(
+                    sub_path,
+                    module_package=None,
+                    package_only=False,
+                )
                 continue
 
-            for suffix in getSharedLibrarySuffixes():
+            for suffix in getExtensionModuleSuffixes():
                 if sub_path.endswith(suffix):
-                    checkPluginSinglePath(sub_path, module_package=None)
+                    scanPluginSinglePath(
+                        sub_path,
+                        module_package=None,
+                        package_only=False,
+                    )
 
     else:
         recursion_logger.warning(
@@ -360,7 +475,7 @@ def checkPluginPath(plugin_filename, module_package):
         )
 
 
-def checkPluginFilenamePattern(pattern):
+def scanPluginFilenamePattern(pattern):
     if Options.isShowInclusion():
         recursion_logger.info("Checking plug-in pattern '%s':" % pattern)
 
@@ -376,7 +491,11 @@ def checkPluginFilenamePattern(pattern):
             continue
 
         found = True
-        checkPluginSinglePath(filename, module_package=None)
+        scanPluginSinglePath(
+            filename,
+            module_package=None,
+            package_only=False,
+        )
 
     if not found:
         recursion_logger.warning(
@@ -384,10 +503,24 @@ def checkPluginFilenamePattern(pattern):
         )
 
 
-def considerUsedModules(module, signal_change):
+def considerUsedModules(module, pass_count):
+    # Modules that are only there because they are in standard library are not
+    # supposed to have dependencies included at all.
+    if module.reason == "stdlib":
+        return
+
     for used_module in module.getUsedModules():
+        # The pass number is used to indicate if stdlib modules yet, or only them
+
+        if used_module.reason == "stdlib":
+            if pass_count == 1:
+                continue
+        else:
+            if pass_count == -1:
+                continue
+
         if used_module.finding == "not-found":
-            Importing.warnAbout(
+            warnAboutNotFoundImport(
                 importing=module,
                 source_ref=used_module.source_ref,
                 module_name=used_module.module_name,
@@ -399,7 +532,7 @@ def considerUsedModules(module, signal_change):
             continue
 
         try:
-            decision, reason = decideRecursion(
+            decision, decision_reason = decideRecursion(
                 using_module_name=module.getFullName(),
                 module_filename=used_module.filename,
                 module_name=used_module.module_name,
@@ -408,20 +541,20 @@ def considerUsedModules(module, signal_change):
 
             if decision:
                 new_module = recurseTo(
-                    signal_change=signal_change,
                     module_name=used_module.module_name,
                     module_filename=used_module.filename,
                     module_kind=used_module.module_kind,
                     source_ref=used_module.source_ref,
+                    reason=used_module.reason,
                     using_module_name=module.module_name,
-                    reason=reason,
                 )
 
                 addUsedModule(
                     module=new_module,
                     using_module=module,
-                    usage_tag="import",
-                    reason=reason,
+                    # TODO: Cleanup argument names here.
+                    usage_tag=used_module.reason,
+                    reason=decision_reason,
                     source_ref=used_module.source_ref,
                 )
         except NuitkaForbiddenImportEncounter as e:
@@ -436,9 +569,43 @@ def considerUsedModules(module, signal_change):
             )
 
     try:
-        Plugins.considerImplicitImports(module=module, signal_change=signal_change)
+        Plugins.considerImplicitImports(module=module)
     except NuitkaForbiddenImportEncounter as e:
         recursion_logger.sysexit(
             "Error, forbidden import of '%s' (intending to avoid '%s') done implicitly by module '%s'."
             % (e.args[0], e.args[1], module.getFullName())
         )
+
+
+def scanIncludedPackage(package_name):
+    package_name, package_directory, _module_kind, finding = locateModule(
+        module_name=ModuleName(package_name),
+        parent_package=None,
+        level=0,
+    )
+
+    if finding != "absolute":
+        recursion_logger.sysexit(
+            "Error, failed to locate package '%s' you asked to include." % package_name
+        )
+
+    scanPluginPath(
+        plugin_filename=package_directory,
+        module_package=package_name.getPackageName(),
+    )
+
+
+#     Part of "Nuitka", an optimizing Python compiler that is compatible and
+#     integrates with CPython, but also works on its own.
+#
+#     Licensed under the Apache License, Version 2.0 (the "License");
+#     you may not use this file except in compliance with the License.
+#     You may obtain a copy of the License at
+#
+#        http://www.apache.org/licenses/LICENSE-2.0
+#
+#     Unless required by applicable law or agreed to in writing, software
+#     distributed under the License is distributed on an "AS IS" BASIS,
+#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+#     See the License for the specific language governing permissions and
+#     limitations under the License.
