@@ -1,4 +1,4 @@
-#     Copyright 2024, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file
+#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file
 
 
 """ Standard plug-in to avoid bloat at compile time.
@@ -14,12 +14,14 @@ import re
 from nuitka.containers.OrderedDicts import OrderedDict
 from nuitka.Errors import NuitkaForbiddenImportEncounter
 from nuitka.ModuleRegistry import getModuleByName
-from nuitka.plugins.PluginBase import NuitkaYamlPluginBase
+from nuitka.Options import isExperimental
+from nuitka.plugins.YamlPluginBase import NuitkaYamlPluginBase
 from nuitka.utils.ModuleNames import ModuleName
 
-# spell-checker: ignore dask,numba,statsmodels,matplotlib,sqlalchemy,ipykernel
+# spell-checker: ignore dask,numba,statsmodels,matplotlib,sqlalchemy,ipykernel,pyximport
 
 _mode_choices = ("error", "warning", "nofollow", "allow")
+_other_choices = ("bytecode",)
 
 
 class NuitkaPluginAntiBloat(NuitkaYamlPluginBase):
@@ -30,6 +32,7 @@ class NuitkaPluginAntiBloat(NuitkaYamlPluginBase):
     plugin_desc = (
         "Patch stupid imports out of widely used library modules source codes."
     )
+    plugin_category = "core"
 
     @staticmethod
     def isAlwaysEnabled():
@@ -40,6 +43,7 @@ class NuitkaPluginAntiBloat(NuitkaYamlPluginBase):
         noinclude_setuptools_mode,
         noinclude_pytest_mode,
         noinclude_unittest_mode,
+        noinclude_pydoc_mode,
         noinclude_ipython_mode,
         noinclude_dask_mode,
         noinclude_numba_mode,
@@ -47,7 +51,8 @@ class NuitkaPluginAntiBloat(NuitkaYamlPluginBase):
         custom_choices,
         show_changes,
     ):
-        # Many details, due to many repetitive arguments, pylint: disable=too-many-branches,too-many-statements
+        # Many details, due to many repetitive arguments,
+        # pylint: disable=too-many-branches,too-many-locals,too-many-statements
 
         NuitkaYamlPluginBase.__init__(self)
 
@@ -60,6 +65,8 @@ class NuitkaPluginAntiBloat(NuitkaYamlPluginBase):
             noinclude_pytest_mode = noinclude_default_mode
         if noinclude_unittest_mode is None:
             noinclude_unittest_mode = noinclude_default_mode
+        if noinclude_pydoc_mode is None:
+            noinclude_pydoc_mode = noinclude_default_mode
         if noinclude_ipython_mode is None:
             noinclude_ipython_mode = noinclude_default_mode
         if noinclude_dask_mode is None:
@@ -102,6 +109,14 @@ class NuitkaPluginAntiBloat(NuitkaYamlPluginBase):
                 noinclude_setuptools_mode,
                 "setuptools",
             )
+            self.handled_modules["numpy.distutils"] = (
+                noinclude_setuptools_mode,
+                "setuptools",
+            )
+            self.handled_modules["wheel.util"] = (
+                noinclude_setuptools_mode,
+                "setuptools",
+            )
         else:
             self.control_tags["use_setuptools"] = True
 
@@ -152,6 +167,7 @@ class NuitkaPluginAntiBloat(NuitkaYamlPluginBase):
         if noinclude_ipython_mode != "allow":
             self.handled_modules["IPython"] = noinclude_ipython_mode, "IPython"
             self.handled_modules["ipykernel"] = noinclude_ipython_mode, "IPython"
+            self.handled_modules["ipywidgets"] = noinclude_ipython_mode, "IPython"
             self.handled_modules["jupyter_client"] = (
                 noinclude_ipython_mode,
                 "IPython",
@@ -173,6 +189,11 @@ class NuitkaPluginAntiBloat(NuitkaYamlPluginBase):
         else:
             self.control_tags["use_dask"] = True
 
+        if noinclude_pydoc_mode != "allow":
+            self.handled_modules["pydoc"] = noinclude_pydoc_mode, "pydoc"
+        else:
+            self.control_tags["use_pydoc"] = True
+
         if noinclude_numba_mode != "allow":
             self.handled_modules["numba"] = noinclude_numba_mode, "numba"
             self.handled_modules["sparse"] = noinclude_numba_mode, "numba"
@@ -188,14 +209,14 @@ class NuitkaPluginAntiBloat(NuitkaYamlPluginBase):
             if custom_choice.count(":") != 1:
                 self.sysexit(
                     """\
-Error, malformed value '%s' for '--noinclude-custom-mode' used. It has to be of \
+Error, argument value '%s' for '--noinclude-custom-mode' used. It has to be of \
 form 'module_name:[%s]'."""
-                    % (custom_choice, "|".join(_mode_choices))
+                    % (custom_choice, "|".join(_mode_choices + _other_choices))
                 )
 
             module_name, mode = custom_choice.rsplit(":", 1)
 
-            if mode not in _mode_choices and mode != "bytecode":
+            if mode not in _mode_choices and mode not in _other_choices:
                 self.sysexit(
                     "Error, illegal mode given '%s' in '--noinclude-custom-mode=%s'"
                     % (mode, custom_choice)
@@ -252,8 +273,9 @@ form 'module_name:[%s]'."""
             action="store_true",
             dest="show_changes",
             default=False,
+            github_action=False,
             help="""\
-Annotate what changes are by the plugin done.""",
+Annotate what changes are done by the plugin.""",
         )
 
         group.add_option(
@@ -289,6 +311,17 @@ What to do if a unittest import is encountered. This package can be big with
 dependencies, and should definitely be avoided.""",
         )
 
+        group.add_option(
+            "--noinclude-pydoc-mode",
+            action="store",
+            dest="noinclude_pydoc_mode",
+            choices=_mode_choices,
+            default=None,
+            help="""\
+What to do if a pydoc import is encountered. This package use is mark of useless
+code for deployments and should be avoided.""",
+        )
+
         group.add_option(
             "--noinclude-IPython-mode",
             action="store",
@@ -367,83 +400,130 @@ Error, cannot exec module '%s', context code '%s' due to: %s"""
 
         return dict(self.context_codes[context_code])
 
-    def _onModuleSourceCode(self, module_name, anti_bloat_config, source_code):
-        # Complex dealing with many cases, pylint: disable=too-many-branches
+    def _onModuleSourceCodeReplacement(
+        self,
+        config_module_name,
+        module_name,
+        source_code,
+        config_name,
+        extra_context,
+        replace_code,
+        replace_src,
+    ):
+        if replace_code:
+            replace_dst = self.evaluateExpression(
+                config_module_name=config_module_name,
+                module_name=module_name,
+                expression=replace_code,
+                config_name="module '%s' config '%s'"
+                % (config_module_name, config_name),
+                extra_context=extra_context,
+                single_value=True,
+            )
+        else:
+            replace_dst = ""
 
+        return source_code, source_code.replace(replace_src, replace_dst)
+
+    def _onModuleSourceCode(
+        self,
+        config_module_name,
+        module_name,
+        config_prefix,
+        anti_bloat_config,
+        source_code,
+    ):
+        # Complex dealing with many cases, pylint: disable=too-many-branches,too-many-locals
         description = anti_bloat_config.get("description", "description not given")
 
         # To allow detection if it did anything.
         change_count = 0
 
-        for replace_src, replace_code in anti_bloat_config.get(
-            "replacements", {}
+        for replace_src, replace_code in (
+            anti_bloat_config.get(config_prefix + "replacements") or {}
         ).items():
             # Avoid the eval, if the replace doesn't hit.
             if replace_src not in source_code:
                 continue
 
-            if replace_code:
-                replace_dst = self.evaluateExpression(
-                    full_name=module_name,
-                    expression=replace_code,
-                    config_name="module '%s' config 'replacements' " % module_name,
+            old, source_code = self._onModuleSourceCodeReplacement(
+                config_module_name=config_module_name,
+                module_name=module_name,
+                source_code=source_code,
+                config_name=config_prefix + "replacements",
+                extra_context=self._getContextCode(
+                    module_name=module_name, anti_bloat_config=anti_bloat_config
+                ),
+                replace_src=replace_src,
+                replace_code=replace_code,
+            )
+
+            if old != source_code:
+                change_count += 1
+
+        for replace_src, replace_dst in (
+            anti_bloat_config.get(config_prefix + "replacements_plain") or {}
+        ).items():
+            old = source_code
+            source_code = source_code.replace(replace_src, replace_dst)
+
+            if old != source_code:
+                change_count += 1
+
+        for replace_src, replace_dst in (
+            anti_bloat_config.get(config_prefix + "replacements_re") or {}
+        ).items():
+            old = source_code
+            source_code = re.sub(replace_src, replace_dst, source_code, re.S)
+
+            if old != source_code:
+                change_count += 1
+            elif isExperimental("display-anti-bloat-mismatches"):
+                self.info("No match in %s no match %r" % (module_name, replace_src))
+
+        # These have no global variants.
+        if config_prefix == "":
+            append_code = anti_bloat_config.get("append_result", "")
+            if type(append_code) in (tuple, list):
+                append_code = "\n".join(append_code)
+
+            if append_code:
+                append_result = self.evaluateExpression(
+                    config_module_name=module_name,
+                    module_name=module_name,
+                    expression=append_code,
+                    config_name="module '%s' config 'append_code'" % module_name,
                     extra_context=self._getContextCode(
                         module_name=module_name, anti_bloat_config=anti_bloat_config
                     ),
                     single_value=True,
                 )
-            else:
-                replace_dst = ""
 
-            old = source_code
-            source_code = source_code.replace(replace_src, replace_dst)
-
-            if old != source_code:
+                source_code += "\n" + append_result
                 change_count += 1
 
-        for replace_src, replace_dst in anti_bloat_config.get(
-            "replacements_plain", {}
-        ).items():
-            old = source_code
-            source_code = source_code.replace(replace_src, replace_dst)
+            append_plain = anti_bloat_config.get("append_plain", "")
+            if type(append_plain) in (tuple, list):
+                append_plain = "\n".join(append_plain)
 
-            if old != source_code:
+            if append_plain:
+                source_code += "\n" + append_plain
                 change_count += 1
 
-        for replace_src, replace_dst in anti_bloat_config.get(
-            "replacements_re", {}
-        ).items():
-            old = source_code
-            source_code = re.sub(replace_src, replace_dst, source_code)
+            module_code = anti_bloat_config.get("module_code", None)
 
-            if old != source_code:
-                change_count += 1
+            if module_code is not None:
+                # Cannot be mixed.
+                assert not change_count
 
-        append_code = anti_bloat_config.get("append_result", "")
-        if type(append_code) in (tuple, list):
-            append_code = "\n".join(append_code)
+                if self.show_changes:
+                    self.info(
+                        "Handling module '%s' with full replacement : %s."
+                        % (module_name.asString(), description)
+                    )
 
-        if append_code:
-            append_result = self.evaluateExpression(
-                full_name=module_name,
-                expression=append_code,
-                config_name="module '%s' config 'append_code'" % module_name,
-                extra_context=self._getContextCode(
-                    module_name=module_name, anti_bloat_config=anti_bloat_config
-                ),
-                single_value=True,
-            )
-
-            source_code += "\n" + append_result
-            change_count += 1
-
-        append_plain = anti_bloat_config.get("append_plain", "")
-        if type(append_plain) in (tuple, list):
-            append_plain = "\n".join(append_plain)
-
-        if append_plain:
-            source_code += "\n" + append_plain
-            change_count += 1
+                source_code = module_code
+                change_count = 1
 
         if change_count > 0 and self.show_changes:
             self.info(
@@ -451,38 +531,55 @@ Error, cannot exec module '%s', context code '%s' due to: %s"""
                 % (module_name.asString(), change_count, description)
             )
 
-        module_code = anti_bloat_config.get("module_code", None)
-
-        if module_code is not None:
-            assert not change_count
-
-            if self.show_changes:
-                self.info(
-                    "Handling module '%s' with full replacement : %s."
-                    % (module_name.asString(), description)
-                )
-
-            source_code = module_code
-
         return source_code
 
     def onModuleSourceCode(self, module_name, source_filename, source_code):
-        for anti_bloat_config in self.config.get(module_name, section="anti-bloat"):
-            if self.evaluateCondition(
-                full_name=module_name, condition=anti_bloat_config.get("when", "True")
+        config_module_name = module_name
+
+        while True:
+            for anti_bloat_config in self.config.get(
+                config_module_name, section="anti-bloat"
             ):
-                source_code = self._onModuleSourceCode(
-                    module_name=module_name,
-                    anti_bloat_config=anti_bloat_config,
-                    source_code=source_code,
-                )
+                if self.evaluateCondition(
+                    full_name=config_module_name,
+                    condition=anti_bloat_config.get("when", "True"),
+                ):
+                    if config_module_name == module_name:
+                        source_code = self._onModuleSourceCode(
+                            config_module_name=config_module_name,
+                            module_name=module_name,
+                            anti_bloat_config=anti_bloat_config,
+                            config_prefix="",
+                            source_code=source_code,
+                        )
+
+                    source_code = self._onModuleSourceCode(
+                        config_module_name=config_module_name,
+                        module_name=module_name,
+                        anti_bloat_config=anti_bloat_config,
+                        config_prefix="global_",
+                        source_code=source_code,
+                    )
+
+            config_module_name = config_module_name.getPackageName()
+
+            if not config_module_name:
+                break
 
         return source_code
 
     def _onFunctionBodyParsing(
-        self, module_name, anti_bloat_config, function_name, body
+        self,
+        config_module_name,
+        module_name,
+        config_prefix,
+        anti_bloat_config,
+        function_name,
+        body,
     ):
-        replace_code = anti_bloat_config.get("change_function", {}).get(function_name)
+        replace_code = anti_bloat_config.get(config_prefix + "change_function", {}).get(
+            function_name
+        )
 
         if replace_code == "un-callable":
             replace_code = """'raise RuntimeError("Must not call %s.%s")'""" % (
@@ -494,10 +591,11 @@ Error, cannot exec module '%s', context code '%s' due to: %s"""
             return False
 
         replacement = self.evaluateExpression(
-            full_name=module_name,
+            config_module_name=config_module_name,
+            module_name=module_name,
             expression=replace_code,
-            config_name="module '%s' config 'change_function' of '%s'"
-            % (module_name, function_name),
+            config_name="module '%s' config '%schange_function' of '%s'"
+            % (module_name, config_prefix, function_name),
             extra_context=self._getContextCode(
                 module_name=module_name, anti_bloat_config=anti_bloat_config
             ),
@@ -529,17 +627,41 @@ Error, cannot exec module '%s', context code '%s' due to: %s"""
     def onFunctionBodyParsing(self, module_name, function_name, body):
         result = False
 
-        config = self.config.get(module_name, section="anti-bloat")
+        config_module_name = module_name
 
-        if config:
-            for anti_bloat_config in config:
-                if self._onFunctionBodyParsing(
-                    module_name=module_name,
-                    anti_bloat_config=anti_bloat_config,
-                    function_name=function_name,
-                    body=body,
+        while True:
+            for anti_bloat_config in self.config.get(
+                config_module_name, section="anti-bloat"
+            ):
+                if self.evaluateCondition(
+                    full_name=config_module_name,
+                    condition=anti_bloat_config.get("when", "True"),
                 ):
-                    result = True
+                    if config_module_name == module_name:
+                        if self._onFunctionBodyParsing(
+                            config_module_name=config_module_name,
+                            module_name=module_name,
+                            anti_bloat_config=anti_bloat_config,
+                            config_prefix="",
+                            function_name=function_name,
+                            body=body,
+                        ):
+                            result = True
+
+                    if self._onFunctionBodyParsing(
+                        config_module_name=config_module_name,
+                        module_name=module_name,
+                        anti_bloat_config=anti_bloat_config,
+                        config_prefix="global_",
+                        function_name=function_name,
+                        body=body,
+                    ):
+                        result = True
+
+            config_module_name = config_module_name.getPackageName()
+
+            if not config_module_name:
+                break
 
         return result
 
@@ -562,7 +684,8 @@ class %(class_name)s:
             return False
 
         replacement = self.evaluateExpression(
-            full_name=module_name,
+            config_module_name=module_name,
+            module_name=module_name,
             expression=replace_code,
             config_name="module '%s' config 'change_class' of '%s'"
             % (module_name, class_name),
@@ -577,7 +700,7 @@ class %(class_name)s:
         replacement = ast.parse(replacement).body[0]
 
         node.body[:] = replacement.body
-        node.bases = replacement.bases
+        node.bases[:] = replacement.bases
 
         if self.show_changes:
             self.info(
@@ -589,10 +712,10 @@ class %(class_name)s:
     def onClassBodyParsing(self, module_name, class_name, node):
         result = False
 
-        config = self.config.get(module_name, section="anti-bloat")
-
-        if config:
-            for anti_bloat_config in config:
+        for anti_bloat_config in self.config.get(module_name, section="anti-bloat"):
+            if self.evaluateCondition(
+                full_name=module_name, condition=anti_bloat_config.get("when", "True")
+            ):
                 if self._onClassBodyParsing(
                     module_name=module_name,
                     anti_bloat_config=anti_bloat_config,
@@ -604,30 +727,62 @@ class %(class_name)s:
         return result
 
     def _getModuleBloatModeOverrides(self, using_module_name, intended_module_name):
-        while 1:
-            config = self.config.get(using_module_name, section="anti-bloat")
+        # Finding a matching configuration aborts the search, not finding one
+        # means default behavior should apply.
+        for _config_module_name, bloat_mode_overrides in self.getYamlConfigItem(
+            module_name=using_module_name,
+            section="anti-bloat",
+            item_name="bloat-mode-overrides",
+            default={},
+            decide_relevant=(lambda config_item: intended_module_name in config_item),
+            recursive=True,
+        ):
+            return bloat_mode_overrides[intended_module_name]
 
-            if config:
-                for anti_bloat_config in config:
-                    bloat_mode_overrides = anti_bloat_config.get(
-                        "bloat-mode-overrides", ()
-                    )
+        return None
 
-                    if not bloat_mode_overrides:
-                        continue
+    def decideAnnotations(self, module_name):
+        # Finding a matching configuration aborts the search, not finding one
+        # means default behavior should apply.
+        for _config_module_name, annotations_config_value in self.getYamlConfigItem(
+            module_name=module_name,
+            section="anti-bloat",
+            item_name="annotations",
+            default=None,
+            decide_relevant=(lambda config_item: config_item in ("yes", "no")),
+            recursive=True,
+        ):
+            return annotations_config_value == "yes"
 
-                    if self.evaluateCondition(
-                        full_name=intended_module_name,
-                        condition=anti_bloat_config.get("when", "True"),
-                    ):
-                        for module_name, mode in bloat_mode_overrides.items():
-                            if module_name == intended_module_name:
-                                return mode
+        return None
 
-            using_module_name = using_module_name.getPackageName()
+    def decideDocStrings(self, module_name):
+        # Finding a matching configuration aborts the search, not finding one
+        # means default behavior should apply.
+        for _config_module_name, doc_strings_config_value in self.getYamlConfigItem(
+            module_name=module_name,
+            section="anti-bloat",
+            item_name="doc_strings",
+            default=None,
+            decide_relevant=(lambda config_item: config_item in ("yes", "no")),
+            recursive=True,
+        ):
+            return doc_strings_config_value == "yes"
 
-            if not using_module_name:
-                break
+        return None
+
+    def decideAsserts(self, module_name):
+        # Finding a matching configuration aborts the search, not finding one
+        # means default behavior should apply.
+        for _config_module_name, asserts_config_value in self.getYamlConfigItem(
+            module_name=module_name,
+            section="anti-bloat",
+            item_name="asserts",
+            default=None,
+            decide_relevant=(lambda config_item: config_item in ("yes", "no")),
+            recursive=True,
+        ):
+            return asserts_config_value == "yes"
 
         return None
 
@@ -645,6 +800,22 @@ class %(class_name)s:
         ):
             self.no_follows[no_follow_pattern] = (config_of_module_name, description)
 
+    def _needsWarning(self, key):
+        if key in self.warnings_given:
+            return False
+
+        (
+            module_name,
+            using_module_name,
+            line_number,
+        ) = key
+
+        for parent_package in module_name.getParentPackageNames():
+            if (parent_package, using_module_name, line_number) in self.warnings_given:
+                return False
+
+        return True
+
     def onModuleRecursion(
         self,
         module_name,
@@ -712,7 +883,7 @@ class %(class_name)s:
                         source_ref.getLineNumber(),
                     )
 
-                    if key not in self.warnings_given:
+                    if self._needsWarning(key):
                         if handled_module_name == intended_module_name:
                             handled_module_name_desc = "'%s'" % handled_module_name
                         else:
@@ -728,7 +899,7 @@ slow down compilation."""
                             % (
                                 handled_module_name_desc,
                                 using_module_name,
-                                source_ref.getAsString(),
+                                self.getReportSourceReference(source_ref),
                             ),
                             mnemonic="unwanted-module",
                         )
@@ -810,6 +981,13 @@ slow down compilation."""
             if module_name.hasNamespace(handled_module_name):
                 return "bytecode"
 
+        # TODO: Detect effective "change_class" and "change_function"
+        # configuration for standard library modules, but often enough we are
+        # happy to let the bytecode not have the effect, but for these ones it's
+        # required. TODO: Make the compilation mode part of the config.
+        if module_name == "xmlrpc.server":
+            return "compiled"
+
     def onModuleCompleteSet(self, module_set):
         # TODO: Maybe have an entry point that works on the set of names
         # instead, we are not looking at the modules, and most plugins probably
